package frc.robot.subsystems;

import com.ctre.phoenix.motorcontrol.can.WPI_VictorSPX;

import edu.wpi.first.math.controller.ElevatorFeedforward;
import edu.wpi.first.math.controller.PIDController;
import edu.wpi.first.util.sendable.SendableBuilder;
import edu.wpi.first.wpilibj.Encoder;
import edu.wpi.first.wpilibj2.command.Command;
import edu.wpi.first.wpilibj2.command.PrintCommand;
import edu.wpi.first.wpilibj2.command.RepeatCommand;
import edu.wpi.first.wpilibj2.command.SubsystemBase;
import edu.wpi.first.wpilibj2.command.WaitCommand;
import frc.robot.constants.ElevatorConstants;

public class Elevator extends SubsystemBase {
    private WPI_VictorSPX motor1;
    private WPI_VictorSPX motor2;

    private Encoder encoder;

    private PIDController positionalController;
    private PIDController velocityController;

    private ElevatorFeedforward eff;

    private double currentBottomPositionalSetpoint;
    private double currentLowMidPositionalSetpoint;
    private double currentHighMidPositionalSetpoint;
    private double currentTopPositionalSetpoint;

    public Elevator() {
        motor1 = new WPI_VictorSPX(ElevatorConstants.kMotor1CANID);
        motor2 = new WPI_VictorSPX(ElevatorConstants.kMotor2CANID);

        encoder = new Encoder(ElevatorConstants.kEncoderChannelADIO, ElevatorConstants.kEncoderChannelBDIO);

        positionalController = new PIDController(
            ElevatorConstants.kPPositional, 
            ElevatorConstants.kIPositional, 
            ElevatorConstants.kDPositional
        );

        velocityController = new PIDController(
            ElevatorConstants.kPVelocity, 
            ElevatorConstants.kIVelocity, 
            ElevatorConstants.kDVelocity
        );

        eff = new ElevatorFeedforward(
            ElevatorConstants.kFFS, 
            ElevatorConstants.kFFG, 
            ElevatorConstants.kFFV
        );

        currentBottomPositionalSetpoint = ElevatorConstants.kDefaultBottomPositionalSetpoint;
        currentLowMidPositionalSetpoint = ElevatorConstants.kDefaultLowMidPositionalSetpoint;
        currentHighMidPositionalSetpoint = ElevatorConstants.kDefaultHighMidPositionalSetpoint;
        currentTopPositionalSetpoint = ElevatorConstants.kDefaultTopPositionalSetpoint;
    }

    public Command debugSetpoints() {
        return new RepeatCommand(
            runOnce(() -> {
                System.out.println("Current Bottom Positional Setpoint: " + currentBottomPositionalSetpoint);
                System.out.println("Current Low-Mid Positional Setpoint: " + currentLowMidPositionalSetpoint);
                System.out.println("Current High-Mid Positional Setpoint: " + currentHighMidPositionalSetpoint);
                System.out.println("Current Top Positional Setpoint: " + currentTopPositionalSetpoint);
                System.out.println("");
            }).andThen(new WaitCommand(5))
        );
    }

    @Override
    public void initSendable(SendableBuilder builder) {
        builder.setActuator(true);

        builder.addDoubleProperty(
            ".currentBottomPositionalSetpoint", 
            () -> currentBottomPositionalSetpoint, 
            (d) -> {
                System.out.println("Current Bottom Positional Setpoint Changed!");
                System.out.println("Old Value: " + currentBottomPositionalSetpoint + " New Value: " + d);
                currentBottomPositionalSetpoint = d;
            }
        );

        builder.addDoubleProperty(
            ".currentLowMidPositionalSetpoint", 
            () -> currentLowMidPositionalSetpoint, 
            (d) -> currentLowMidPositionalSetpoint = d
        );

        builder.addDoubleProperty(
            ".currentHighMidPositionalSetpoint", 
            () -> currentHighMidPositionalSetpoint, 
            (d) -> currentHighMidPositionalSetpoint = d
        );

        builder.addDoubleProperty(
            ".currentTopPositionalSetpoint", 
            () -> currentTopPositionalSetpoint, 
            (d) -> currentTopPositionalSetpoint = d
        );
    }
}
